'''
Created on 2022-05-15

@author: wf
'''

class Version(object):
    '''
    Version handling for pyJustpyWidgets
    '''
    name="pyJustpyWidgets"
    version = '0.1.8'  
    date = '2022-05-15'
    updated = '2022-11-27'
    description='justpy Widgets'
